#!/usr/bin/bash
# ==============================================================================
# Migration package
# ==============================================================================
# This package contains the functionalities needed for the transfer of data from
# the original Latin-1 DB to the new UTF-8 DB. This needs to be installed in the
# target database and run from there.
# Before you run this script, first check the configuration section at the very
# top of it and adjust the values to reflect your environment.
#
# Author: AI Rehberg
# $Id: iz_migration.sh,v 1.5 2007-02-09 16:28:42 izzy Exp $

# ==========================================================[ Configuration ]===
# Name of the database link to the originating DB, as well as user/pwd to log on
# there:
DBLINK=mp15.devoracle
DBLINK_USER=system
DBLINK_PASS=oracle24
# Datafile of the tablespace holding the objects needed by our package
DATAFILE=/u01/oradata/mutf8/iz_migration.dbf
# Where to log the object creation
SPOOL=iz_migration.log
# Binary to execute - normally this is sqlplus, but to simply generate the
# code you can also use 'cat' and redirect the output to a SQL file
#BINARY='cat'
BINARY='sqlplus "/as sysdba"'

# ========================================================[ Install Objects ]===
$BINARY<<EOF
SPOOL $SPOOL
SET ECHO ON FEEDBACK ON DEFINE OFF FLUSH ON NEWPAGE NONE PAGESIZE 0 TERMOUT ON SERVEROUTPUT ON

-- First we create a new tablespace where the iz_migration objects will be stored
CREATE TABLESPACE iz_migration DATAFILE
  '$DATAFILE' SIZE 10M AUTOEXTEND ON NEXT 10M MAXSIZE UNLIMITED
LOGGING
PERMANENT
EXTENT MANAGEMENT LOCAL UNIFORM SIZE 1M;

-- The user running the migration is:
CREATE USER "IZ_MIGRATION" IDENTIFIED BY "oracle"
 DEFAULT TABLESPACE iz_migration
 QUOTA UNLIMITED ON iz_migration;
GRANT CONNECT TO IZ_MIGRATION;
GRANT DBA TO IZ_MIGRATION;
GRANT CREATE PROCEDURE TO IZ_MIGRATION;
GRANT SELECT_CATALOG_ROLE TO IZ_MIGRATION;
GRANT SELECT ANY DICTIONARY TO IZ_MIGRATION;
GRANT SYSDBA TO IZ_MIGRATION;
-- Some explicit permissions are needed for PL/SQL
GRANT ALTER ANY INDEX TO IZ_MIGRATION;
GRANT ALTER ANY SEQUENCE TO IZ_MIGRATION;
GRANT ALTER ANY TABLE TO IZ_MIGRATION;
GRANT ALTER ANY TRIGGER TO IZ_MIGRATION;
GRANT CREATE ANY INDEX TO IZ_MIGRATION;
GRANT CREATE ANY SEQUENCE TO IZ_MIGRATION;
GRANT CREATE ANY TABLE TO IZ_MIGRATION;
GRANT CREATE ANY TRIGGER TO IZ_MIGRATION;
GRANT DROP ANY SEQENCE TO IZ_MIGRATION;
GRANT EXECUTE ANY PROCEDURE TO IZ_MIGRATION;
GRANT SELECT ANY DICTIONARY TO IZ_MIGRATION;
GRANT SELECT ANY SEQUENCE TO IZ_MIGRATION;
GRANT SELECT ANY TABLE TO IZ_MIGRATION;
GRANT UPDATE ANY TABLE TO IZ_MIGRATION;

-- Now we sitch to the migration user to create the needed objects
connect iz_migration/oracle

-- Create the DBLink needed to transfer the data
CREATE DATABASE LINK "$DBLINK"
CONNECT TO $DBLINK_USER
IDENTIFIED BY "$DBLINK_PASS"
USING '$DBLINK';

-- Create the table holding the names of the schemata to xfer
-- and insert all names of the schemata
CREATE TABLE iz_owners (name VARCHAR2(30));

-- Table for the logging messages
CREATE TABLE iz_logging (   -- here we log all our actions
  id         NUMBER,        -- for sorting purposes
  log_time   DATE,          -- what time the entry was made
  log_level  CHAR(1),       -- here: "!" for errors, "*" for main (job start/end),
                            -- "+" for sub (e.g. switch to next owner), "-" for
                            -- minor
  log_entry  VARCHAR2(4000) -- the log message
);
CREATE SEQUENCE IZ_MIGRATION.SEQ_LOGGING -- Sequence for the Logging tables ID column
  START WITH 0 MINVALUE 0 INCREMENT BY 1
  CACHE 20 NOCYCLE ORDER;
CREATE OR REPLACE TRIGGER IZ_MIGRATION.trig_logging_id -- Trigger to auto-increment iz_logging.id
  BEFORE INSERT ON IZ_MIGRATION.IZ_LOGGING
  REFERENCING NEW AS New OLD AS Old
  FOR EACH ROW
  DECLARE
    tmpVar NUMBER;
  BEGIN
   tmpVar := 0;
   SELECT seq_logging.NEXTVAL INTO tmpVar FROM dual;
   :NEW.id := tmpVar;
   EXCEPTION
     WHEN OTHERS THEN
       RAISE;
END trig_logging_id;
/

-- Save the original state of all constraints
-- Other than with the triggers, we can't use the DB link here, since "unnamed"
-- constraint names (SYS_*) will be freshly generated on the import and thus
-- differ from the original ones. These are usually "NOT NULL" constraints...
CREATE TABLE iz_constraints AS
 SELECT owner,constraint_name,constraint_type,table_name,status,validated
   FROM dba_constraints
  WHERE 0=1;

-- Tables for the semantics conversion
CREATE TABLE iz_semantics (
  s_owner VARCHAR(40),
  s_table_name VARCHAR(40),
  s_column_name VARCHAR(40),
  s_data_type VARCHAR(40),
  s_char_length NUMBER
);
CREATE TABLE iz_chars AS SELECT * FROM iz_semantics WHERE 0=1;
CREATE TABLE iz_varchars AS SELECT * FROM iz_semantics WHERE 0=1;
CREATE TABLE iz_clobs AS SELECT * FROM iz_semantics WHERE 0=1;

-- Table to backup oversized character data
CREATE TABLE iz_long_fields (
  s_owner VARCHAR(40),
  s_table_name VARCHAR(40),
  s_column_name VARCHAR(40),
  s_row_id ROWID,
  s_data CLOB
);

-- Package definition for our transfer procedures
CREATE OR REPLACE PACKAGE iz_migration AUTHID CURRENT_USER AS
  PROCEDURE print(line IN VARCHAR2); -- wrapper to dbms_output.put_line
  PROCEDURE log(level IN CHAR,message IN VARCHAR2); -- generate log entries to iz_logging
  PROCEDURE dry_run (enabled IN BOOLEAN); -- Set DryRun mode: TRUE: Only log statements, FALSE: execute statements
  PROCEDURE enable_trigger; -- enable all triggers which have been enabled before
  PROCEDURE disable_trigger; -- disable all triggers which are enabled in this DB
  PROCEDURE enable_constraints; -- enable all constraints which have been enabled before
  PROCEDURE disable_constraints; -- disable all constraints which are enabled in this DB
  PROCEDURE validate_constraints; -- validate all enabled constraints for the schemata listed in the iz_owners table
  PROCEDURE truncate_tables; -- truncate all tables for owners defined in the iz_owners table
  PROCEDURE drop_indexes; -- drop all non-essential indices (i.e. no-PK/unique/LOB/IOT) for re-start
  PROCEDURE copy_tables; -- copy all tables for owners defined in the iz_owners table
  PROCEDURE copy_sequences; -- drop and recreate all sequences for owners defined in iz_owners
  PROCEDURE semantics2char; -- migrate the semantics from byte to char for all user-tables in the DB
  PROCEDURE initiate_semantics_table; -- Trucate and refill table holding specifications for conversion objects
  PROCEDURE initiate_nchars_table; -- Truncate and refill table for NCHAR to CHAR conversion
  PROCEDURE initiate_nvarchars_table; -- Truncate and refill table for NVARCHAR2 to VARCHAR2 conversion
  PROCEDURE initiate_nclobs_table; -- Truncate and refill table for NCLOB to CLOB conversion
  PROCEDURE initiate_constraints_table; -- filling the table with constraint information
  PROCEDURE convert_nchar; -- Convert all NCHAR to CHAR
  PROCEDURE convert_nvarchar; -- Convert all NVARCHAR2 to VARCHAR2
  PROCEDURE convert_nclob; -- Convert all NCLOB to CLOB
  PROCEDURE truncate_long_fields; -- after conversion to UTF-8, some content might be consuming more than 4000 bytes
  PROCEDURE schema_stats(s_percent IN NUMBER := 30,s_degree IN NUMBER := 4); -- Generating statistics for all schemata listed in iz_owners
  PROCEDURE compile_invalid; -- Compile all procedures, functions, packages & triggers for the iz_owners
END iz_migration;
/

-- Package body for our transfer procedures
CREATE OR REPLACE PACKAGE BODY iz_migration AS
  dryrun BOOLEAN := FALSE;
  PROCEDURE print(line IN VARCHAR2) IS
    -- wrapper to dbms_output.put_line - work around the 255 characters limit
    pos NUMBER;
    FUNCTION strpos (str IN VARCHAR2,needle IN VARCHAR2,startpos NUMBER) RETURN NUMBER IS
      -- split string helper to procedure print
      pos NUMBER; strsub VARCHAR2(255);
      BEGIN
        strsub := SUBSTR(str,1,255);
        pos    := INSTR(strsub,needle,startpos);
        return pos;
      END;
    BEGIN
      dbms_output.put_line(line);
    EXCEPTION
      WHEN OTHERS THEN
        IF SQLERRM LIKE '%ORU-10028%' THEN
          pos := strpos(line,' ',-1);
	  print(SUBSTR(line,1,pos));
	  pos := pos +1;
	  print(SUBSTR(line,pos));
	ELSE
          dbms_output.put_line('-- *!* Problem in print() *!*');
	END IF;
    END;
  PROCEDURE log(level IN CHAR,message IN VARCHAR2) IS
    -- Write log message to the logging table. If this fails, write to
    -- dbms_output using the print() wrapper.
    -- Parameters:
    --   level   - one of '!' (error), '*' (job start/end), '+' (subjob start/end)
    --             or '-' (notice)
    --   message - log message
    msg VARCHAR2(4000);
    BEGIN
      IF LENGTHB(message)>4000 THEN
        msg := substr(message,1,4000); -- some SQL statements might be quite long
      ELSE
        msg := message;
      END IF;
      IF LENGTH(level)>1 THEN
        log('!','Procedure log() called with invalid first parameter (more than 1 char)');
        log('#',msg);
      ELSE
        INSERT INTO iz_logging (log_time,log_level,log_entry) VALUES (SYSDATE,level,msg);
        COMMIT;
      END IF;
    EXCEPTION
      WHEN OTHERS THEN
        print('! Logging failed: '||SQLERRM);
        print(level||' '||to_char(SYSDATE,'YYYY-MM-DD HH24:MI')||' '||message);
    END;
  PROCEDURE dry_run (enabled IN BOOLEAN) IS
    BEGIN
      IF enabled THEN
        log('*','Enabling DryRun');
      ELSE
        log('*','Disabling DryRun');
      END IF;
      dryrun := enabled;
    END;
  PROCEDURE exec_sql(statement IN VARCHAR2) IS
    cur INTEGER;
    row_nums NUMBER;
    BEGIN
      IF dryrun THEN
        log('S',statement);
      ELSE
        cur := dbms_sql.open_cursor;
        dbms_sql.parse(cur,statement,dbms_sql.native);
        row_nums := dbms_sql.execute(cur);
        COMMIT;
        dbms_sql.close_cursor(cur);
      END IF;
    END;
  PROCEDURE enable_trigger IS
    -- Enable all triggers for the schemata listed in the iz_owners table
    CURSOR c_trigger IS
      SELECT owner,trigger_name,table_owner,table_name FROM dba_triggers@$DBLINK
       WHERE status='ENABLED' AND owner IN (SELECT name FROM iz_owners);
    BEGIN
      log('*','Enabling triggers...');
      FOR rec IN c_trigger LOOP
        BEGIN
          exec_sql('ALTER TRIGGER '||rec.owner||'.'||rec.trigger_name||' ENABLE');
          exec_sql('ALTER TRIGGER '||rec.owner||'.'||rec.trigger_name||' COMPILE');
        EXCEPTION
          WHEN OTHERS THEN log('!','ERROR enabling trigger: '||SQLERRM);
        END;
      END LOOP;
      log('*','Triggers enabled.');
    EXCEPTION
      WHEN OTHERS THEN
        log('!','ERROR enabling trigger: '||SQLERRM);
    END;
  PROCEDURE disable_trigger IS
    -- Disable all triggers for the schemata listed in the iz_owners table
    CURSOR c_trigger IS
      SELECT owner,trigger_name,table_owner,table_name FROM dba_triggers@$DBLINK
       WHERE status='ENABLED' AND owner IN (SELECT name FROM iz_owners);
    BEGIN
      log('*','Disabling triggers...');
      FOR rec IN c_trigger LOOP
        BEGIN
          exec_sql('ALTER TRIGGER '||rec.owner||'.'||rec.trigger_name||' DISABLE');
	EXCEPTION
          WHEN OTHERS THEN log('!','ERROR disabling trigger: '||SQLERRM);
	END;
      END LOOP;
      log('*','Triggers disabled.');
    EXCEPTION
      WHEN OTHERS THEN
        log('!','ERROR disabling trigger: '||SQLERRM);
    END;
  PROCEDURE enable_constraints IS
    -- Enable all constraints for the schemata listed in the iz_owners table
    tctype CHAR(1) := '0';
    CURSOR c_constraints IS
      SELECT owner,table_name,constraint_name,constraint_type FROM iz_constraints
       WHERE status='ENABLED' AND owner IN (SELECT name FROM iz_owners)
       ORDER BY DECODE(constraint_type,'U',1,'P',2,'R',3,'C',4,5); -- resolve dependencies: 1st Unique, 2nd PK, 3rd all others
    PROCEDURE do_enable(towner IN VARCHAR2,tname IN VARCHAR2,cname IN VARCHAR2) IS
      BEGIN
        exec_sql('ALTER TABLE '||towner||'.'||tname||' MODIFY CONSTRAINT '||cname||' ENABLE NOVALIDATE');
      EXCEPTION
        WHEN OTHERS THEN log('!','ERROR enabling constraint '||cname||' on table '||towner||'.'||tname||': '||SQLERRM);
      END;
    BEGIN
      log('*','Enabling constraints...');
      FOR rec IN c_constraints LOOP
        IF tctype != rec.constraint_type THEN
          tctype := rec.constraint_type;
          log('+','Enabling constraints of type '||tctype);
        END IF;
        do_enable(rec.owner,rec.table_name,rec.constraint_name);
      END LOOP;
      log('*','Constraints enabled.');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR in enable_constraints: '||SQLERRM);
    END;
  PROCEDURE validate_constraints IS
    -- Validate all enabled constraints for the schemata listed in the iz_owners table
    counter NUMBER :=0;
    oname VARCHAR2(30) := '';
    CURSOR c_constraints IS
      SELECT owner,table_name,constraint_name FROM dba_constraints
       WHERE status='ENABLED' AND validated!='VALIDATED' AND owner IN (SELECT name FROM iz_owners)
       ORDER BY owner;
    PROCEDURE do_validate(tname IN VARCHAR2, cname IN VARCHAR2) IS
      BEGIN
        log('-','Validating constraint "'||cname||'" for table "'||tname||'"');
        exec_sql('ALTER TABLE '||tname||' MODIFY CONSTRAINT '||cname||' VALIDATE');
        counter := counter +1;
      EXCEPTION
        WHEN OTHERS THEN log('!','Could not validate constraint '||cname||' on table '||tname||': '||SQLERRM);
      END;
    BEGIN
      log('*','Validating constraints...');
      FOR rec IN c_constraints LOOP
        IF rec.owner!=oname THEN
          oname := rec.owner;
          log('+','Validating constraints for "'||oname||'"');
        END IF;
        do_validate(rec.owner||'.'||rec.table_name,rec.constraint_name);
      END LOOP;
      log('*',counter||' constraints validated.');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR in enable_constraints: '||SQLERRM);
    END;
  PROCEDURE disable_constraints IS
    -- Disable all constraints for the schemata listed in the iz_owners table
    errors NUMBER := 0;
    CURSOR c_constraints IS
      SELECT owner,table_name,constraint_name FROM iz_constraints
       WHERE status='ENABLED' AND owner IN (SELECT name FROM iz_owners)
       ORDER BY DECODE(constraint_type,'R',1,'P',2,3); -- resolve dependencies: 1st FK, 2nd PK, 3rd all others
    BEGIN
      log('*','Disabling constraints...');
      FOR rec IN c_constraints LOOP
        BEGIN
          exec_sql('ALTER TABLE '||rec.owner||'.'||rec.table_name||' MODIFY CONSTRAINT '||rec.constraint_name||' DISABLE');
        EXCEPTION
          WHEN OTHERS THEN
            errors := errors +1;
            log('!','ERROR disabling constraint '||rec.owner||'.'||rec.constraint_name||': '||SQLERRM);
        END;
      END LOOP;
      log('*','Constraints disabled ('||errors||' error(s)).');
    EXCEPTION
      WHEN OTHERS THEN
        log('!','ERROR in disable_constraints: '||SQLERRM);
    END;
  PROCEDURE copy_sequences IS
    -- Re-Create all Sequences for the schemata listed in the iz_owners table
    -- using last_number from the source database
    lastnum NUMBER;
    CURSOR c_sequences IS
      SELECT sequence_owner,sequence_name,last_number FROM dba_sequences@$DBLINK
	   WHERE sequence_owner IN (SELECT name FROM iz_owners);
    BEGIN
      log('*','Recreating Sequences...');
      FOR rec IN c_sequences LOOP
        exec_sql('DROP SEQUENCE '||rec.sequence_owner||'.'||rec.sequence_name);
        exec_sql('CREATE SEQUENCE '||rec.sequence_owner||'.'||rec.sequence_name
                        ||' START WITH '||rec.last_number);
      END LOOP;
      log('*','Sequences recreated.');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR in copy_sequences: '||SQLERRM);
    END;
  PROCEDURE truncate_tables IS
    -- Copy data of all tables for the schemata listed in the iz_owners table
    tname VARCHAR2(30) := '0';
    errors NUMBER := 0;
    CURSOR c_tables IS
	  SELECT owner,table_name FROM dba_tables
	   WHERE owner IN (SELECT name FROM iz_owners)
	     AND table_name NOT LIKE 'RUPD$%' AND table_name NOT LIKE 'MLOG$%'
	     AND temporary!='Y'
           ORDER BY owner,table_name;
    PROCEDURE truncate_table(towner IN VARCHAR2,tname IN VARCHAR2) IS
      BEGIN
        exec_sql('TRUNCATE TABLE '||towner||'.'||tname);
      EXCEPTION
        WHEN OTHERS THEN
          errors := errors +1;
          log('!','ERROR truncating table '||tname||': '||SQLERRM);
      END;
    BEGIN
      log('*','Truncating tables...');
      FOR rec IN c_tables LOOP
        IF tname != rec.owner THEN
          tname := rec.owner;
          log('+','Truncating tables for '||tname);
        END IF;
        log('-','Truncate '||tname||'.'||rec.table_name);
        truncate_table(rec.owner,rec.table_name);
      END LOOP;
      log('*','Truncation finished with '||errors||' error(s).');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR truncating table data: '||SQLERRM);
    END;
  PROCEDURE drop_indexes IS
    counter NUMBER := 0;
    CURSOR cur IS
      SELECT i.owner,i.index_name
        FROM dba_indexes i,dba_constraints c
       WHERE i.owner = c.owner(+)
         AND i.index_name = c.index_name(+)
         AND c.index_name IS NULL
         AND i.index_type NOT LIKE 'IOT%'
         AND i.index_type != 'LOB'
         AND i.owner IN (SELECT name FROM iz_owners)
       ORDER BY i.owner,i.index_name;
    PROCEDURE drop_idx(iname IN VARCHAR2) IS
      BEGIN
        exec_sql('DROP INDEX '||iname);
        counter := counter +1;
      EXCEPTION
        WHEN OTHERS THEN log('!','ERROR dropping index "'||iname||'": '||SQLERRM);
      END;
    BEGIN
      log('*','Dropping indexes');
      FOR rec IN cur LOOP
        drop_idx(rec.owner||'.'||rec.index_name);
      END LOOP;
      log('*','Finished dropping '||counter||' indexes.');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR dropping indexes: '||SQLERRM);
    END;
  PROCEDURE copy_tables IS
    -- Copy data of all tables for the schemata listed in the iz_owners table
    tname VARCHAR2(30) := '0';
    counter NUMBER :=0;
    CURSOR c_tables IS
	  SELECT owner,table_name FROM dba_tables
	   WHERE owner IN (SELECT name FROM iz_owners)
	     AND table_name NOT LIKE 'RUPD$%' AND table_name NOT LIKE 'MLOG$%'
	     AND table_name!='STATMEMBER' AND table_name!='REPORTARCHIVE'
	     AND temporary!='Y'
           ORDER BY owner,table_name;
    PROCEDURE copy_table(towner IN VARCHAR2,tname IN VARCHAR2) IS
      BEGIN
        exec_sql('ALTER TABLE '||towner||'.'||tname||' NOLOGGING');
        exec_sql('INSERT /*+ APPEND */ INTO '||towner||'.'||tname||' SELECT * FROM '||towner||'.'||tname||'@$DBLINK');
        exec_sql('ALTER TABLE '||towner||'.'||tname||' LOGGING');
        counter := counter +1;
      EXCEPTION
        WHEN OTHERS THEN log('!','ERROR copying table '||tname||': '||SQLERRM);
      END;
    BEGIN
      log('*','Copying tables...');
      FOR rec IN c_tables LOOP
        IF tname != rec.owner THEN
          tname := rec.owner;
          log('+','Copying tables for '||tname);
        END IF;
        log('-','Copy '||tname||'.'||rec.table_name);
        copy_table(rec.owner,rec.table_name);
      END LOOP;
      log('*','Finished copying '||counter||' tables.');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR copying table data: '||SQLERRM);
    END;
  PROCEDURE semantics2char IS
    -- migrate the semantics from byte to char for all user-tables in the DB
    CURSOR c_sem IS SELECT * FROM iz_semantics;
    statement VARCHAR2(255);
    numcols NUMBER(10);
    numtabs NUMBER(10);
    town VARCHAR(40) := '0';
    BEGIN
      EXECUTE IMMEDIATE 'SELECT COUNT(*) FROM iz_semantics' INTO numcols;
      EXECUTE IMMEDIATE 'SELECT COUNT(DISTINCT s_table_name) FROM iz_semantics' INTO numtabs;
      log('*','Changing semantics for '||numcols||' columns in '||numtabs||' tables');
      FOR rec IN c_sem LOOP
        IF town != rec.s_owner THEN
          town := rec.s_owner;
          log('+','ALTERing objects of user '||town);
        END IF;
        log('-','ALTERing semantic of table '||rec.s_owner||'.'||rec.s_table_name);
        statement := 'ALTER TABLE '||rec.s_owner||'.'||rec.s_table_name
                  || ' MODIFY ('||rec.s_column_name||' '||rec.s_data_type
                  || '('||rec.s_char_length||' CHAR))';
        exec_sql(statement);
      END LOOP;
      log('*','All semantics changed to BYTE');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR changing semantics: '||SQLERRM);
    END;
  PROCEDURE convert_national(tname IN VARCHAR2,datatype IN VARCHAR2) IS
    -- Convert the N*CHAR* fields into *CHAR*
    cur SYS_REFCURSOR;
    statement VARCHAR2(255);
    numcols NUMBER(10);
    numtabs NUMBER(10);
    town VARCHAR(40) := '';
    s_owner VARCHAR2(40);
    s_table_name VARCHAR2(40);
    s_column_name VARCHAR2(40);
    s_char_length NUMBER;
    BEGIN
      EXECUTE IMMEDIATE 'SELECT COUNT(*) FROM iz_'||tname INTO numcols;
      EXECUTE IMMEDIATE 'SELECT COUNT(DISTINCT s_table_name) FROM iz_'||tname INTO numtabs;
      log('*',datatype||' changing for '||numcols||' columns in '||numtabs||' tables');
      statement := 'SELECT s_owner,s_table_name,s_column_name,s_char_length FROM iz_'||tname;
      OPEN cur FOR statement;
      LOOP
        FETCH cur INTO s_owner,s_table_name,s_column_name,s_char_length;
        EXIT WHEN cur%NOTFOUND;
        IF town != s_owner THEN
          town := s_owner;
          log('+','ALTERing objects of user '||town);
        END IF;
        log('-','ALTERing semantic of table '||s_owner||'.'||s_table_name);
        statement := 'ALTER TABLE '||s_owner||'.'||s_table_name
                  || ' MODIFY ('||s_column_name||' '||datatype
                  || '('||s_char_length||' CHAR))';
        exec_sql(statement);
      END LOOP;
      CLOSE cur;
      log('*','All changed to '||datatype||' using BYTE semantic');
    EXCEPTION
      WHEN OTHERS THEN
        IF cur%ISOPEN THEN
          CLOSE cur;
        END IF;
        log('!','ERROR changing data type '||datatype||': '||SQLERRM);
    END;
  PROCEDURE convert_nchar IS
    BEGIN
      convert_national('chars','CHAR');
    END;
  PROCEDURE convert_nvarchar IS
    BEGIN
      convert_national('varchars','VARCHAR2');
    END;
  PROCEDURE convert_nclob IS
    BEGIN
      convert_national('clobs','CLOB');
    END;
  PROCEDURE initiate_table (tname IN VARCHAR2,datatypes IN VARCHAR2) IS
    -- Trucate and refill tables holding specifications for conversion objects
    statement VARCHAR(8000);
    s_owner VARCHAR2(40);
    s_table_name VARCHAR2(40);
    s_column_name VARCHAR2(40);
    s_data_type VARCHAR2(40);
    s_char_length NUMBER;
    counter NUMBER := 0;
    cur SYS_REFCURSOR;
    BEGIN
      log('*','Initiate table for '||tname);
      exec_sql('TRUNCATE TABLE iz_'||tname);
      statement := '  SELECT C.owner, C.table_name, C.column_name, C.data_type, C.char_length'
                || '    FROM all_tab_columns C, all_tables T'
                || '   WHERE C.owner = T.owner'
                || '     AND T.owner NOT IN (''SYS'',''SYSTEM'',''CSMIG'',''CTXSYS'',''DBSNMP'',''DMSYS'','
                || '                         ''EXFSYS'',''HR'',''IX'',''MDSYS'',''OE'',''OLAPSYS'','
                || '                         ''ORDSYS'',''OUTLN'',''SH'',''SYSMAN'',''WKSYS'','
                || '                         ''WK_TEST'',''WMSYS'',''XDB'')'
                || '     AND C.table_name = T.table_name'
                || '     AND C.table_name NOT IN (SELECT table_name FROM all_external_tables)'
                || '     AND C.data_type IN ('||datatypes||')';
      open cur for statement;
      LOOP
        FETCH cur INTO s_owner,s_table_name,s_column_name,s_data_type,s_char_length;
        EXIT WHEN cur%NOTFOUND;
        statement := 'INSERT INTO iz_'||tname||' VALUES ('''||s_owner||''','''||s_table_name||''','''||s_column_name
                  || ''','''||s_data_type||''','||s_char_length||')';
        exec_sql(statement);
        counter := counter +1;
      END LOOP;
      CLOSE cur;
      log('*','Table for '||tname||' initiated with '||counter||' rows.');
      COMMIT;
    EXCEPTION
      WHEN OTHERS THEN
        IF cur%ISOPEN THEN
          CLOSE cur;
        END IF;
        log('!','ERROR initiating table for '||tname||': '||SQLERRM);
    END;
  PROCEDURE initiate_semantics_table IS
    BEGIN
      initiate_table('semantics','''VARCHAR2'',''CHAR''');
    END;
  PROCEDURE initiate_nchars_table IS
    -- Truncate and refill table for NCHAR to CHAR conversion
    BEGIN
      initiate_table('chars','''NCHAR''');
    END;
  PROCEDURE initiate_nvarchars_table IS
    -- Truncate and refill table for NVARCHAR2 to CHAR conversion
    BEGIN
      initiate_table('varchars','''NVARCHAR2''');
    END;
  PROCEDURE initiate_nclobs_table IS
    -- Truncate and refill table for NCHAR to CHAR conversion
    BEGIN
      initiate_table('clobs','''NCLOB''');
    END;
  PROCEDURE initiate_constraints_table IS
    statement VARCHAR2(8000);
    counter NUMBER := 0;
    CURSOR cur IS
      SELECT owner,constraint_name,constraint_type,table_name,status,validated
        FROM dba_constraints
       WHERE owner IN (SELECT name FROM iz_owners)
       ORDER BY owner;
    BEGIN
      log('*','Initializing constraints table');
      exec_sql('TRUNCATE TABLE iz_constraints');
      FOR rec IN cur LOOP
        statement := 'INSERT INTO iz_constraints VALUES ('''||rec.owner||''','''
                  || rec.constraint_name||''','''||rec.constraint_type||''','''
                  || rec.table_name||''','''||rec.status||''','''||rec.validated||''')';
        exec_sql(statement);
        counter := counter +1;
      END LOOP;
      log('*','Constraints table initiated with '||counter||' constraints.');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR initiating constraints table: '||SQLERRM);
    END;

  PROCEDURE schema_stats(s_percent IN NUMBER := 30,s_degree IN NUMBER := 4) IS
    -- Generate statistics for all schemata we transfered data into
    town VARCHAR2(40); -- needed for exception handler
    CURSOR cur IS
	  SELECT name FROM iz_owners;
    CURSOR cur2 (towner IN VARCHAR2) IS
      SELECT table_name,iot_type FROM dba_tables WHERE owner=towner ORDER BY table_name;
    BEGIN
      log('*','Generating statistics...');
      FOR rec IN cur LOOP
        town := rec.name;
        log('+','Gathering statistics for '||rec.name);
        dbms_stats.delete_schema_stats(ownname=>rec.name); -- drop old stats first
        BEGIN
          dbms_stats.gather_schema_stats(ownname=>rec.name,estimate_percent=>s_percent,
                                         cascade=>TRUE,degree=>s_degree);
        EXCEPTION
          WHEN OTHERS THEN
            IF SQLERRM LIKE '%ORA-00600%' THEN
              log('!','ORA-00600 while gathering stats for '||rec.name||'. Going table-wise.');
              FOR rec2 IN cur2(rec.name) LOOP
                BEGIN
                  log('-','Gathering stats for '||rec.name||'.'||rec2.table_name);
                  dbms_stats.gather_table_stats(ownname=>rec.name,estimate_percent=>s_percent,
                                                cascade=>TRUE,tabname=>rec2.table_name);
                EXCEPTION
                  WHEN OTHERS THEN
                    log('!','Failed gathering statistics for '||rec.name||'.'||rec2.table_name||' - try ANALYZE as last resort');
                    BEGIN
                      EXECUTE IMMEDIATE 'ANALYZE TABLE '||rec.name||'.'||rec2.table_name||' ESTIMATE STATISTICS';
                      IF rec2.iot_type!='IOT' THEN
                        EXECUTE IMMEDIATE 'ANALYZE TABLE '||rec.name||'.'||rec2.table_name||' ESTIMATE STATISTICS FOR ALL INDEXES';
                      END IF;
                    EXCEPTION
                      WHEN OTHERS THEN
                        log('!','ANALYZE failed as well for '||rec.name||'.'||rec2.table_name||', sorry - you need to care manually.');
                    END;
                END;
              END LOOP;
            ELSIF SQLERRM LIKE '%ORA-20000%' THEN
              log('!','Schema '||town||' does not exist (or insufficient privileges to analyze it)');
            END IF;
        END;
      END LOOP;
      log('*','Statistics generated.');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR generating statistics for '||town||': '||SQLERRM);
    END;
  PROCEDURE truncate_long_fields IS
    -- *CHAR field contents may consume more than the maximum of 4000 bytes after migration
    s_owner   VARCHAR2(40) := '';
    s_table   VARCHAR2(40) := '';
    s_column  VARCHAR2(4000) := '';
    s_row_id  ROWID;
    s_count   NUMBER;
    tstring   VARCHAR2(8000);
    statement VARCHAR2(8000);
    refcur SYS_REFCURSOR;
    CURSOR cur IS
      SELECT *
        FROM iz_semantics
       WHERE s_owner IN (SELECT name FROM iz_owners)
         AND s_char_length>1999
       ORDER BY s_owner,s_table_name,s_column_name;
    BEGIN
      log('*','Truncating fields which have more than 4000 Bytes after conversion');
      FOR rec IN cur LOOP
        IF (s_owner!=rec.s_owner) THEN
          log('+','Checking schema '||rec.s_owner);
          s_owner := rec.s_owner;
        END IF;
        IF (s_table!=rec.s_table_name) THEN
          s_table := rec.s_table_name;
          log('-','Checking table '||rec.s_owner||'.'||rec.s_table_name);
        END IF;
        log('','Checking field '||rec.s_owner||'.'||rec.s_table_name||'.'||rec.s_column_name);
        statement := 'SELECT rowid AS row_id,'||rec.s_column_name||' FROM '||rec.s_owner||'.'||rec.s_table_name||'@$DBLINK'
                  || ' WHERE LENGTH('||rec.s_column_name||')>1999';
        BEGIN
          OPEN refcur FOR statement;
          LOOP
            FETCH refcur INTO s_row_id,s_column;
            EXIT WHEN refcur%NOTFOUND;
            tstring := CONVERT(s_column,'AL32UTF8','WE8ISO8859P15');
            IF LENGTHB(tstring)>4000 THEN
              s_count := 0;
              WHILE LENGTHB(CONVERT(s_column,'AL32UTF8','WE8ISO8859P15')) > 4000 LOOP
                s_column := substr(s_column,1,LENGTH(s_column) -1);
                s_count := s_count +1;
              END LOOP;
              log('$','RowID '''||s_row_id||''' of '||rec.s_owner||'.'||rec.s_table_name||'.'||rec.s_column_name||' must be truncated by '||s_count||' characters');
              statement := 'INSERT INTO iz_long_fields VALUES ('''||rec.s_owner||''','''||rec.s_table_name||''','''||rec.s_column_name
                        || ''','''||s_row_id||''','''||REPLACE(s_column,'''','''''')||''')';
              exec_sql(statement);
              statement := 'UPDATE '||rec.s_owner||'.'||rec.s_table_name||'@$DBLINK SET '||rec.s_column_name||'='''||REPLACE(s_column,'''','''''')
                        || ''' WHERE rowid='''||s_row_id||'''';
              exec_sql(statement);
            END IF;
          END LOOP;
          CLOSE refcur;
        EXCEPTION
          WHEN OTHERS THEN
            IF refcur%ISOPEN THEN
              CLOSE refcur;
            END IF;
            log('!','ERROR in truncation process: '||SQLERRM);
        END;
      END LOOP;
      log('*','Truncation process finished.');
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR truncating long fields: '||SQLERRM);
    END;

  PROCEDURE compile_invalid IS
    -- Compile all procedures, functions, packages & triggers for the iz_owners
    CURSOR cur IS
      SELECT name FROM iz_owners;
    BEGIN
      FOR rec IN cur LOOP
        dbms_utility.compile_schema(schema=>rec.name);
      END LOOP;
    EXCEPTION
      WHEN OTHERS THEN log('!','ERROR compiling objects - '||SQLERRM);
    END;

END iz_migration;
/
COMMIT;
SPOOL OFF
EOF

